package com.agentfactory.agentspeak.interpreter.mentalstate;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.agentspeak.interpreter.Utilities;
import com.agentfactory.logic.update.lang.ITerm;
import com.agentfactory.logic.update.lang.Variable;
import com.agentfactory.logic.update.reasoner.Bindings;

public class Commitment {
	private ASEvent event;
	private List<IPlanStep> stack;
	private IPlanStep current;
	private int state;
	private List<Bindings> bindings = new LinkedList<Bindings>();
	private List<Integer> popPoints = new LinkedList<Integer>();
	private Map<Integer, List<IPlanStep>> recoveryPlans = new HashMap<Integer, List<IPlanStep>>();
	
	public static final int STATE_ACTIVE						 = 0;
	public static final int STATE_SUSPENDED						 = 1;
	public static final int STATE_FAILED						 = 2;

	
	public Commitment(ASEvent event, PlanRule primary) {
		this.event = event;
		stack = new LinkedList<IPlanStep>();
		addPlan(primary);
		bindings.add(new Bindings());
		popPoints.add(0);
	}

	public String toString() {
		String output = event.toString() + ": ";
		for (Bindings b: bindings) {
			output += Utilities.presenter.toString(b) + " / ";
		}
		return output;
	}
	
	public ASEvent getEvent() {
		return event;
	}
	
	public List<IPlanStep> getStack() {
		return stack;
	}
	
	public IPlanStep getCurrent() {
		return current;
	}
	
	public void addPlan(PlanRule rule) {
		List<IPlanStep> subplan = rule.getPlan();
		for (int i=subplan.size()-1; i>=0; i--) {
			stack.add(0, subplan.get(i));
		}
	}
	
	public void addPlan(List<IPlanStep> subplan) {
		pushBindings();
		popPoints.add(stack.size());
		for (int i=subplan.size()-1; i>=0; i--) {
			stack.add(0, subplan.get(i));
		}
	}
	
	public void back() {
		stack.add(0, current);
	}
	
	public IPlanStep getNextActivity() {
		if (stack.isEmpty()) {
			current = null;
		} else {
			if (popPoints.get(popPoints.size()-1) == stack.size()) {
				popPoints.remove(popPoints.size()-1);
				popBindings();
			}
			current = stack.remove(0);
		}
		
		return current;
	}
	
	public void suspend() {
		state = Commitment.STATE_SUSPENDED;
	}
	
	public void resume() {
		state = Commitment.STATE_ACTIVE;
	}
	
	public void fail() {
		state = Commitment.STATE_FAILED;
	}
	
	public boolean isSuspended() {
		return state == Commitment.STATE_SUSPENDED;
	}
	
	public boolean isActive() {
		return state == Commitment.STATE_ACTIVE;
	}

	public boolean isFailed() {
		return state == Commitment.STATE_FAILED;
	}
	
	public void pushBindings() {
		bindings.add(new Bindings());
	}
	
	public void popBindings() {
		bindings.remove(bindings.size()-1);
	}
	
	public void updateBindings(Bindings bindings) {
		int current = this.bindings.size() - 1;
		this.bindings.set(current, this.bindings.get(current).merge(bindings));
	}
	
	public void updateBinding(Variable variable, ITerm term) {
		int index = 0;
		while (index < bindings.size()) {
			Bindings b = bindings.get(index++);
			ITerm current = b.getBinding(variable);
			if (current != null) {
				b.addBinding(variable, term);
				return;
			}
		}
		
		bindings.get(bindings.size()-1).addBinding(variable, term);
	}
	
	public Bindings bindings() {
		Bindings current = new Bindings();
		for (Bindings b:bindings) {
			current = current.merge(b);
		}
		return current;
	}

	/**
	 * Adds a recovery plan to the commitment. This plan is linked to the stack size when it
	 * is added. The plan is triggered in the case where the commitment has failed and the
	 * stack is being rolled back. When the stack level corresponds to a recovery plan, recovery
	 * occurs.
	 * 
	 * @param plan
	 */
	public void addRecoveryPlan(List<IPlanStep> plan) {
		recoveryPlans.put(stack.size(), plan);
	}
	
	public boolean rollback() {
		while (!stack.isEmpty() && (recoveryPlans.get(stack.size()) == null)) {
			getNextActivity();
		}

		if (stack.isEmpty()) return false;

		popBindings();
		addPlan(recoveryPlans.remove(stack.size()));
		
		this.resume();
		
		return true;
	}
}
