package com.agentfactory.logic.update.reasoner;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import com.agentfactory.logic.update.interpreter.CoreUtilities;
import com.agentfactory.logic.update.lang.Equality;
import com.agentfactory.logic.update.lang.Function;
import com.agentfactory.logic.update.lang.GreaterThan;
import com.agentfactory.logic.update.lang.IFormula;
import com.agentfactory.logic.update.lang.IMPLIES;
import com.agentfactory.logic.update.lang.ITerm;
import com.agentfactory.logic.update.lang.LessThan;
import com.agentfactory.logic.update.lang.NOT;
import com.agentfactory.logic.update.lang.Predicate;
import com.agentfactory.logic.update.lang.Unifier;
import com.agentfactory.logic.update.lang.Variable;

public class ResolutionBasedQueryEngine implements IQueryEngine {
    public static int MAX_LEVELS                                    = 20;
    private static int counter = 0;

    private List<IQueryable> sources;
    private List<StackEntry> possibleSolutions;
    private Set<Variable> variableSet;
    private List<StackEntry> completedSet;
    
    class StackEntry {
        List<IFormula> goalStack;
        Bindings solution;
        int level;
        
        private StackEntry() {
            goalStack = new LinkedList<IFormula>();
            solution = new Bindings();
            level = 0;
        }
        
        public StackEntry(IFormula goal) {
            this();
            addGoal(goal);
        }
        
        @Override
        public String toString() {
            StringBuffer buf = new StringBuffer();
            buf.append("goalStack: ");
            for (IFormula fos:goalStack) {
                buf.append(CoreUtilities.presenter.toString(fos));
                buf.append(" ");
            }
            buf.append(" / solution: ");
            buf.append(CoreUtilities.presenter.toString(solution));
            buf.append(" / level: " + level);
            return buf.toString();
        }
        
        public StackEntry copy(Bindings set) {
            StackEntry entry = new StackEntry();
            
            List<IFormula> list = new LinkedList<IFormula>();
            for (IFormula goal: goalStack) {
            	list.add(goal.apply(new Bindings()));
            }
            entry.goalStack = list;
            entry.solution = solution.merge(set).idempotencyBuilder().mostGeneralUnifier();
            entry.level = level+1;
            return entry;
        }
        
        public void addGoal(IFormula formula) {
            level++;
            for (IFormula f: formula.asList()) {
            	goalStack.add(f);
            }
        }
        
        public IFormula getNextGoal() {
            return goalStack.remove(0).apply(solution);
        }
        
        /**
         * remove all variables that were not in the initial goal.
         * 
         * @param variableSet
         */
        public void update(Set<Variable> variableSet) {
        	solution = solution.filter(variableSet);
        }
    }
    
	public ResolutionBasedQueryEngine() {
		sources = new LinkedList<IQueryable>();
	}
	
	public void addSource(IQueryable source) {
		sources.add(source);
	}

    private List<IFormula> getMatchingFormula(IFormula formula) {
        List<IFormula> formulae = new LinkedList<IFormula>();
        for (IQueryable source:sources) {
            if (source.hasType(formula.type())) {
                formulae.addAll( source.query(formula) );
            }
        }
        return formulae;
    }

    private void setup(IFormula object) {
        completedSet = new LinkedList<StackEntry>();
        possibleSolutions = new LinkedList<StackEntry>();
        possibleSolutions.add(new StackEntry(object));
        variableSet = object.getVariables();
    }
    
    public QueryResult query(IFormula formula) {
    	QueryResult result = new QueryResult();
    	result.found = doQuery(formula, true);
    	
    	if (completedSet != null) {
	        for (StackEntry entry: completedSet) {
	        	if (!result.bindings.contains(entry.solution)) result.bindings.add(entry.solution);
	        }
    	}
        return result;
    }

    public QueryResult queryAll(IFormula formula) {
    	QueryResult result = new QueryResult();
    	result.found = doQuery(formula, false);
    	
    	if (completedSet != null) {
	        for (StackEntry entry: completedSet) {
	        	if (!result.bindings.contains(entry.solution)) result.bindings.add(entry.solution);
	        }
    	}
        return result;
    }

    
    /**
     * Take a set of current beliefs and belief rules, and resolve all the
     * beliefs that are entailed.
     *
     * The only difference between this and the above, is two thirds in.
     *
     * @param goal - A Query Object. That which is to be checked for
     * entailment.
     * @return boolean - true if inBeliefManager entails inQuery, false otherwise.
     */
    private boolean doQuery(IFormula goal, boolean stopOnSolution) {
        // Setup the reasoner -- when complete, possibleSolutions will have one
        // entry with the goal and no substitution.
        setup(goal);
        
        // Go through possibilities until the solution stack is empty.
        while(!possibleSolutions.isEmpty()) {
            StackEntry solution = possibleSolutions.remove(0);
            
            if ( solution.goalStack.isEmpty() ) {
                // The potential solution has no goals left to be solved. This
                // potential solution must be a valid solution.
                completedSet.add(solution);
            	if (stopOnSolution) {
            		// if we only want to know whether or not there is a solution,
            		// then simply break from the loop at this point...
            		break;
            	}
            } else if (solution.level < MAX_LEVELS) {
                // lets take this potential solution and try to work it out
                // If this is successful, a stack entry with no goals will be
                // added to the list of possible solutions. Else, if something
                // needs further investigation, a stack entry with goals as well
                // as a substitution set will be added to the set of possibilities.
                // Else, this solution was a dead end and nothing is added to
                // the set of possibilities.
                
                // Also, we check here whether the maximum search depth (MAX_LEVELS)
                // has been reached - if so, then drop the potential solution...
                solve(solution);
            }
        }
        
        for(int i=0; i < this.completedSet.size(); i++) {
            completedSet.get(i).update(variableSet);
        }
        
        return !(completedSet.isEmpty());
    }

    /**
     * Apply a resolution step to the current solution
     * @param entry
     */
    private void solve(StackEntry entry) {
        // Reform a fully formed goal with applications of whatever substitutions
    	// are appropriate
        IFormula formula = entry.getNextGoal().apply(entry.solution);
        
        if(NOT.class.isInstance(formula)) {
            if(!spawn().doQuery(((NOT) formula).formula(), true)) {
                possibleSolutions.add(entry);
            }
            return;
        } else if (!Predicate.class.isInstance(formula)) {
        	if (Equality.class.isInstance(formula) || LessThan.class.isInstance(formula) || GreaterThan.class.isInstance(formula)) {
        		if (formula.isTrue()) possibleSolutions.add(entry);
        		return;
        	}
        	throw new RuntimeException("QueryEngine Error: all goals should be predicates, but have: " + formula);
        }

        Predicate goal = (Predicate) formula;
//        System.out.println("Checking: " + CoreUtilities.presenter.toString(goal));
        if (goal.isTrue()) {
            possibleSolutions.add(entry);
        } else if (!goal.isFalse()) {
            if (comparisonCheck(goal)) {
                possibleSolutions.add(entry.copy(new Bindings()));
            } else {
                List<IFormula> matchingFormulae = getMatchingFormula(goal);
                for(IFormula nextMatch:matchingFormulae) {
                    // Check what type of object this is and process accordingly.
                    if (IMPLIES.class.isInstance(nextMatch)) {
                        IMPLIES implication = (IMPLIES) nextMatch;
                        Bindings b1 = variableRenamingFor(implication);
                        // Pre unification check to confirm annotations match
                        Predicate result = (Predicate) implication.result().apply(b1);
                        generateLabelBindings(goal, result); 
                        Bindings b2 = Unifier.unifyFormulae(goal,result);
                        // if subs2 != null, we were able to unify the head of the
                        // belief rule with the current goal.
                        if(b2 != null) {
                            // make a copy of the main Stack Entry adding the tail
                            // of the belief rule -- giving us more to work out
                            StackEntry stackEntry = entry.copy(b2);
                            stackEntry.addGoal( implication.condition().apply(b1).apply(b2) );
                            possibleSolutions.add(stackEntry);
                        }
                    } else if (Predicate.class.isInstance(nextMatch)) {
//                		System.out.println("test: " + CoreUtilities.presenter.toString(nextMatch));
                        List<Bindings> labelBindings = generateLabelBindings(goal, (Predicate) nextMatch);
                        if (!labelBindings.isEmpty()) {
                        	for (Bindings binding : labelBindings) {
                        		Bindings b1 = Unifier.unifyFormulae(goal, (Predicate) nextMatch.apply(binding));
//                        		System.out.println("b1: " + CoreUtilities.presenter.toString(b1));
                        		if(b1 != null) {
                        			StackEntry se = entry.copy(b1.merge(binding));
                        			possibleSolutions.add(se);
                        		}
                        	}
                        } else if (goal.getLabels().isEmpty()) {
                    		Bindings b1 = Unifier.unifyFormulae(goal, (Predicate) nextMatch);
                    		if(b1 != null) {
                    			StackEntry se = entry.copy(b1);
                    			possibleSolutions.add(se);
                    		}
                        }
                    } else {
                        throw new RuntimeException("Fatal Error in Query Engine: attempt to match: " + nextMatch);
                    }
                }
            }
        }
    }
    
    private List<Bindings> generateLabelBindings(Predicate goal, Predicate result) {
    	List<Bindings> bindings = new LinkedList<Bindings>();
    	
    	for (Predicate label: goal.getLabels()) {
    		for (Predicate rLabel: result.getLabels()) {
    			if (label.matches(rLabel)) {
    				Bindings binding = Unifier.unifyFormulae(label, rLabel);
    				if (binding != null) {
    					bindings.add(binding);
    				}
    			}
    		}
    	}
    	
    	return bindings;
    }
    
    /**
     * Generates a SubstitutionSet that contains a set of bindings for
     * variable renaming.
     *
     * @param object the formula to be renamed.
     * @return a substitution set containing the renamings.
     */
    private static Bindings variableRenamingFor(IFormula formula) {
        Bindings subs = new Bindings();
        Set<Variable> variables = formula.getVariables();
        for (Variable variable: variables) {
        	subs.addBinding(variable, new Variable(variable.functor() + counter));
        }
        counter++;
        return subs;
    }
    
    /**
     * Checks various comparison operators.
     * TODO (Rem Collier): Need to refactor this to use new logic model...
     * 
     * @param goal
     * @return
     */
    private boolean comparisonCheck(IFormula formula) {
        if (!Predicate.class.isInstance(formula)) {
        	return false;
        }
        
        Predicate predicate = (Predicate) formula;
    	String operator = predicate.functor();
        
        if (operator.equals("IN_RANGE")) {
            try {
                int target = Integer.parseInt(predicate.termAt(0).toString());
                int lower = Integer.parseInt(predicate.termAt(1).toString());
                int upper = Integer.parseInt(predicate.termAt(2).toString());
                return (target >= lower) && (target <= upper);
            } catch (NumberFormatException nfe) {
                return false;
            }
        } else if (operator.equals("CONTAINS")) {
            ITerm target = predicate.termAt(0);
            ITerm list = predicate.termAt(1);
            if (!Function.class.isInstance(list)) {
            	return false;
            }
            for (ITerm value: (Function) list) {
                if (value.matches(target)) {
                    return true;
                }
            }
        }
            
        return false;
    }

    private ResolutionBasedQueryEngine spawn() {
        ResolutionBasedQueryEngine engine = new ResolutionBasedQueryEngine();
        for (IQueryable source: sources) {
            engine.addSource(source);
        }
        return engine;
    }

    public void setMaxDepth(int max){
    	MAX_LEVELS = max;
    }
}
