package com.agentfactory.logic.update.util;

import com.agentfactory.logic.update.lang.AND;
import com.agentfactory.logic.update.lang.Addition;
import com.agentfactory.logic.update.lang.Division;
import com.agentfactory.logic.update.lang.Equality;
import com.agentfactory.logic.update.lang.Function;
import com.agentfactory.logic.update.lang.GreaterThan;
import com.agentfactory.logic.update.lang.Head;
import com.agentfactory.logic.update.lang.IFormula;
import com.agentfactory.logic.update.lang.IMPLIES;
import com.agentfactory.logic.update.lang.ITerm;
import com.agentfactory.logic.update.lang.IntegerTerm;
import com.agentfactory.logic.update.lang.LessThan;
import com.agentfactory.logic.update.lang.ListTerm;
import com.agentfactory.logic.update.lang.Merge;
import com.agentfactory.logic.update.lang.Modulo;
import com.agentfactory.logic.update.lang.Multiplication;
import com.agentfactory.logic.update.lang.NOT;
import com.agentfactory.logic.update.lang.Predicate;
import com.agentfactory.logic.update.lang.RealTerm;
import com.agentfactory.logic.update.lang.Subtraction;
import com.agentfactory.logic.update.lang.Tail;
import com.agentfactory.logic.update.lang.Variable;
import com.agentfactory.logic.update.reasoner.Bindings;

/**
 * This class implements support for generating string representations
 * of logical formulae that is based on the default logic syntax used in
 * Agent Factory:
 * - and (&)
 * - variable (?)
 * - not (~)
 * - implication (=>)
 * 
 * To use a different syntax all you need to do is create a new Logic
 * Presenter and a new Logic Factory implementation.
 * 
 * To extend the basic syntax to include modal/temporal operators, ...
 * simply subclass this class and override the toString(IFormula ...)
 * method. The overriding method should call the superclass method
 * to generate string representations of the basic syntax. For example: 
 * 
 * public String toString(IFormula formula) {
 *   if (MyCustomFormula.class.isInstance(formula)) {
 *     .. generate and return string
 *   }
 *   
 *   return super.toString(formula);
 * }
 * 
 * @author rem
 *
 */
public class DefaultLogicPresenter implements ILogicPresenter {

	public String toString(Bindings bindings) {
		String output = "{";
		boolean first = true;
		for (ITerm[] terms: bindings) {
			if (first) first = false;
			else output += ",";
			output += toString(terms[0]) + "/" + toString(terms[1]);
		}
		return output + "}";
	}
	
	@Override
	public String toString(IFormula formula) {
		if (Predicate.class.isInstance(formula)) {
			Predicate predicate = (Predicate) formula;
			String output = predicate.functor();
			
			if (!predicate.isConstant()) {
				output += "(";
				boolean first = true;
				for (ITerm t: predicate) {
					if (first) first = false;
					else output += ",";
					output += toString(t);
				}
				output += ")";
			}
			if (!predicate.getLabels().isEmpty()) {
				output += "[";
				boolean first = true;
				for (Predicate label : predicate.getLabels()) {
					if (first) first = false;
					else output += ",";
					output += toString(label);
				}
				output += "]";
			}
			return output;
		} else if (NOT.class.isInstance(formula)) {
			return "~" + toString(((NOT) formula).formula());
		} else if (Equality.class.isInstance(formula)) {
			Equality eq = (Equality) formula;
			return toString(eq.left()) + "==" + toString(eq.right());
		} else if (GreaterThan.class.isInstance(formula)) {
			GreaterThan gt = (GreaterThan) formula;
			return toString(gt.left()) + ">" + toString(gt.right());
		} else if (LessThan.class.isInstance(formula)) {
			LessThan lt = (LessThan) formula;
			return toString(lt.left()) + "<" + toString(lt.right());
		} else if (AND.class.isInstance(formula)) {
			String output = "";
			boolean first = true;
			for (IFormula f:(AND) formula) {
				if (first) first = false;
				else output += "&";
				output += toString(f);
			}
			return output;
		} else if (IMPLIES.class.isInstance(formula)) {
			IMPLIES implies = (IMPLIES) formula;
			return toString(implies.condition()) + " => " + toString(implies.result()); 
		}
		
		return null;
	}

	@Override
	public String toString(ITerm term) {
		if (Variable.class.isInstance(term)) {
			return "?" + ((Variable) term).functor();
		} else if (ListTerm.class.isInstance(term)) {
			ListTerm list = (ListTerm) term;
			
			// Remove the quotations on constant functions
			String output = "[";
			boolean first = true;
			for (ITerm t: list) {
				if (first) first = false;
				else output += ",";
				output += toString(t);
			}
			
			return output + "]";
		} else if (Head.class.isInstance(term)) {
			return "head(" + toString(((Head) term).getTerm()) + ")";
		} else if (Tail.class.isInstance(term)) {
			return "tail(" + toString(((Tail) term).getTerm()) + ")";
		} else if (Merge.class.isInstance(term)) {
			Merge merge = (Merge) term;
			return "merge(" + toString(merge.term(0)) + "," + toString(merge.term(1)) + ")";
		} else if (Function.class.isInstance(term)) {
			Function function = (Function) term;
			
			// Remove the quotations on constant functions
			String output = function.functor();
			if (output.startsWith("\"") && output.endsWith("\"")) {
				output = output.substring(1,output.length()-1);
			}
			
			if (function.isConstant()) {
				return output;
			}
			
			output += "(";
			boolean first = true;
			for (ITerm t: function) {
				if (first) first = false;
				else output += ",";
				output += toString(t);
			}
			
			return output + ")";
		} else if (Modulo.class.isInstance(term)) {
			Modulo add = (Modulo) term;
			return toString(add.term(0)) + " % " + toString(add.term(1)); 
		} else if (Addition.class.isInstance(term)) {
			Addition add = (Addition) term;
			return toString(add.term(0)) + " + " + toString(add.term(1)); 
		} else if (Subtraction.class.isInstance(term)) {
			Subtraction add = (Subtraction) term;
			return toString(add.term(0)) + " - " + toString(add.term(1)); 
		} else if (Multiplication.class.isInstance(term)) {
			Multiplication add = (Multiplication) term;
			return toString(add.term(0)) + " * " + toString(add.term(1)); 
		} else if (Division.class.isInstance(term)) {
			Division add = (Division) term;
			return toString(add.term(0)) + " / " + toString(add.term(1)); 
		} else if (IntegerTerm.class.isInstance(term)) {
			return ""+((IntegerTerm) term).value(); 
		} else if (RealTerm.class.isInstance(term)) {
			return ""+((RealTerm) term).value(); 
		}
		return null;
	}

	public static void main(String[] args) {
		DefaultLogicFactory factory = new DefaultLogicFactory();
		DefaultLogicPresenter presenter = new DefaultLogicPresenter();
		
		IFormula formula = factory.createFormula("happy(Howell)[source(rem), source(percepts), confidence(0.5)]");
		System.out.println(presenter.toString(formula));
	}
}
