/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package com.agentfactory.platform.impl;

import com.agentfactory.platform.core.IAgent;
import com.agentfactory.platform.core.IScheduler;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author remcollier
 */
public final class RoundRobinTimeSliceScheduler extends Thread implements IScheduler {
    private static String LOGNAME = RoundRobinTimeSliceScheduler.class.getName();
    
    static {
        Logger.getLogger(LOGNAME).setLevel(Level.SEVERE);
    }
    
    private Queue<IAgent> scheduleQueue;
    private Map<String, IAgent> agentMap;
    private Map<String, AgentThread> threadMap;
    private boolean terminated;
    private IAgent active;
    
    public RoundRobinTimeSliceScheduler() {
        scheduleQueue = new LinkedList<IAgent>();
        threadMap = new HashMap<String, AgentThread>();
        agentMap = new HashMap<String, IAgent>();
        terminated = false;
        setName("Scheduler Thread");
        setPriority(7);
        start();
    }
    
    /**
     * Registers the agent with the scheduler for execution.
     * 
     * @param agent the agent to be scheduled
     */
    public synchronized void register(IAgent agent) {
        AgentThread agentThread = new AgentThread(agent);
        Thread thread = new Thread(agentThread, agent.getName());
        threadMap.put(agent.getName(), agentThread);
        agentMap.put(agent.getName(), agent);
        scheduleQueue.add(agent);
        thread.setPriority(5);
        thread.start();
    }

    /**
     * Removed the given agent from the schedule.
     * 
     * @param agent the agent to be un-scheduled
     */
    public synchronized void unregister(IAgent agent) {
        agentMap.remove(agent.getName());
        threadMap.remove(agent.getName());
        
        // Agent will be terminated at the start of the next execution...
        agent.setState(IAgent.TERMINATED);
    }

    /**
     * Resume the given agent
     * 
     * @param name
     */
    public void resume(String name) {
        IAgent agent = agentMap.get(name);
        agent.setState(IAgent.ACTIVE);
    }

    /**
     * Suspend the given agent
     * 
     * @param name
     */
    public void suspend(String name) {
        IAgent agent = agentMap.get(name);
        agent.setState(IAgent.SUSPENDED);
    }

    /**
     * Step the given agent. This method is synchronised so that the agent
     * cannot be executed multiple times (concurrently).
     * 
     * @param name
     */
    public void step(String name) {
        IAgent agent = agentMap.get(name);
        agent.setState(IAgent.SUSPENDED);
        agent.execute();
    }
    
    /**
     * Check if the given agent is currently active (i.e. it has been resumed)
     * 
     * @param agent
     * @return
     */
    private boolean isActive(IAgent agent) {
        return agent.getState() == IAgent.ACTIVE;
    }
    
    /**
     * This is the main control thread for the scheduling system. Currently,
     * each agent is scheduled sequentially based on a round-robin approach.
     * Agents that are inactive are removed from the head of the queue and
     * immediately returned to the end of the queue (i.e. they are not
     * executed).
     */
    @Override
    public void run() {
        while (!isTerminated()) {
            active = scheduleQueue.peek();
            if (active != null) {
                if (active.getState() == IAgent.ACTIVE) {
                    boolean interrupted = false;
                    
                    // Get the agent thread and invoke it...
                    AgentThread agentThread = threadMap.get(active.getName());
                    active.setEndOfTimeSlice(false);
                    
                    long startTime = System.currentTimeMillis();
                    agentThread.interrupt();
                    
                    try {
                        Thread.sleep( active.getTimeSlice() );
                    } catch ( InterruptedException ie ) {
                        interrupted = true;
                    }
                    
                    if ( !interrupted ) {
                        Logger.getLogger(RoundRobinTimeSliceScheduler.class.getName()).
                                log(Level.INFO, "[Scheduler] Agent: " + active.getName() + " did not finish on time.");
                        active.setEndOfTimeSlice(true);
                        try {
                            synchronized ( this ) {
                                wait();
                            }
                        } catch (InterruptedException ie) {
                            Logger.getLogger(RoundRobinTimeSliceScheduler.class.getName()).
                                    log(Level.INFO, "[Scheduler] Agent: " + active.getName() + " recieved end of time slice.");
                        }
                    }
                    long duration = System.currentTimeMillis() - startTime;

                    Logger.getLogger(RoundRobinTimeSliceScheduler.class.getName()).
                            log(Level.INFO, "[Scheduler] Agent: " + active.getName() + " took: " + duration + " milliseconds.");
                    
                }
            
                // Move the agent to the back of the queue
                scheduleQueue.offer(scheduleQueue.remove());
            }
            
            // Sleep for a short period
            try {
                Thread.sleep(10);
            } catch (InterruptedException ex) {
                Logger.getLogger(RoundRobinTimeSliceScheduler.class.getName()).log(Level.SEVERE, null, ex);
            }
            
        }
    }
    
    private synchronized boolean isTerminated() {
        return terminated;
    }
    
    public synchronized void terminate() {
        terminated = true;
    }
    
    /**
     * Notify the scheduler that the agent has finished executing
     */
    public synchronized void finishedExecuting() {
        this.interrupt();
    }

    /**
     * Implementation of Agent Control Thread
     */
    private class AgentThread implements Runnable {
        IAgent agent;
        
        public AgentThread(IAgent agent) {
            this.agent = agent;
        }

        public void run() {
            do {
                doWait();
                agent.execute();
                finishedExecuting();
            } while (!isTerminated());
        }
        
        private synchronized void doWait() {
            try {
                wait();
            } catch (InterruptedException ex) {
                Logger.getLogger(RoundRobinTimeSliceScheduler.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        
        private synchronized boolean isTerminated() {
            return agent.getState() == IAgent.TERMINATED;
        }
        
        public synchronized void interrupt() {
            notify();
        }
    }
}
